import numpy as np
import numpy.random as random
import matplotlib.pyplot as plt

# slow-learning regime: p_u = 1/1000 theta = 1/150 p = 0.30
# fast-learning regime: p_u = 1/250 theta = 1/75 p = 0.10

J = 12 # number of marketbits
N = 25 # number of investors
M = 99 # number of non-default predictors
S = N # number of shares
nrun = 110000

p = 0.1 # probability for crossover over mutation in GA
p_u = 1./250. # learning rate (probability for implementing asynchronous GA
r = 0.10 # risk-free interest rate
c = 0.005 # fitness coefficient for specificity
gamma = 0.5 # risk averseness coefficient of agents' CARA utility function
theta = 1./75. # moving average coefficient for predictors' updating of actual variance

market_state = [0]*J

price = [100.]
dividend = [10.]
volume = [0.]
hree_price = [100.]
specificities = [[0., 0., 0.]]

time_series = [price, volume, hree_price, dividend]
time_series_labels = ['Price', 'Volume', 'HREE.Price', 'Dividend']

class Predictor ():

    def __init__(self):
        self.condition = []
        self.initialize_condition()
        self.action= [random.uniform(.7,1.2),random.uniform(-10.,19.002)]
        self.actual_variance = 4.
        self.variance = 4.
        self.fitness = - self.variance - c*len([bit for bit in self.condition if bit != -1])
        self.last_active = 0

    def initialize_condition (self):
        for i in range(J):
            if (random.random() < .9):
                self.condition.append(-1)
            else:
                if (random.random() < .5):
                    self.condition.append(1)
                else:
                    self.condition.append(0)

    def compare_market_state():
        signal = True
        for i,bit in enumerate(market_state):
            if (self.condition[i] != bit and self.condition[i] != -1):
                signal = False
                break
        return signal

    def mutate(self, parent):
        # condition mutation
        self.condition = parent.condition
        for i,bit in enumerate(self.condition):
            if (np.random.random() < .03):
                if (bit == 0):
                    if (random.random() < 2./3.):
                        self.condition[i] = -1
                    else:
                        self.condition[i] = 1
                if (bit == 1):
                    if (random.random() < 2./3.):
                        self.condition[i] = -1
                    else:
                        self.condition[i] = 0
                if (bit == -1):
                    if (random.random() < 1./3.):
                        self.condition[i] = 0
                    else:
                        if (np.random.random < .5):
                            self.condition[i] = 1
                        else:
                            self.condition[i] = -1
        # action mutation
        if random.random() < .6:
            self.action = parent.action
        else:    
            if (np.random.random() < .5):
                self.action = [np.random.uniform(.7,1.2),np.random.uniform(-10.,19.002)]
            else:
                self.action = [np.random.uniform(.95-0.00025,.95+0.00025),np.random.uniform(4.501-0.014501,4.501+0.014501)]
        # variance, fitness and last_active initialization
        self.variance = parent.variance
        self.actual_variance = self.variance
        self.fitness = - self.variance - 0.005*len([bit for bit in self.condition if bit != -1])
        self.last_active = parent.last_active

    def recombine(self, parent_1, parent_2, predictors):
        # condition recombination
        for i in range(len(self.condition)):
            if (np.random.random() < .5):
                self.condition[i] = parent_1.condition[i]
            else:
                self.condition[i] = parent_2.condition[i]
        # action recombination
        if (np.random.random() < 1./3.):
            if (np.random.random() < .5):
                self.action[0] = parent_1.action[0]
                self.action[1] = parent_2.action[1]
            else:
                self.action[0] = parent_2.action[0]
                self.action[1] = parent_1.action[1]
        else:
            if (np.random.random() < .5):
                a = ((1./parent_1.variance)*parent_1.action[0] /
                    + (1./parent_2.variance)*parent_2.action[0])/(1./parent_1.variance+1./parent_2.variance)
                b = ((1./parent_1.variance)*parent_1.action[1] /
                    + (1./parent_2.variance)*parent_2.action[1])/(1./parent_1.variance+1./parent_2.variance)
                self.action = [a, b]
            else:
                if (np.random.random() < .5):
                    self.action = parent_1.action
                else:
                    self.action = parent_2.action
        # variance recombination
        if (parent_1.last_active > 500 or parent_2.last_active > 500):
            self.variance = np.mean([parent_1.variance, parent_2.variance])
        else:
            self.variance = np.median([predictor.variance for predictor in predictors])
        # fitness and last_active initialization
        self.actual_variance = self.variance
        self.fitness = - self.variance - 0.005*len([bit for bit in self.condition if bit != -1])
        self.last_active = np.max([parent_1.last_active, parent_2.last_active])
        
    def mutate_for_inactivity(self):
        bits_index = [i for i in range(J) if self.condition[i] != -1]
        np.random.shuffle(bits_index)
        i = 0
        while (i+1 <= len(bits_index)/4.):
            self.condition[bits_index[i]] = -1
            i += 1

class Investor ():
    
    def __init__(self):
        self.predictors = []
        for j in range(M):
            predictor = Predictor()
            self.predictors.append(predictor)
        self.active_predictors = []
        self.last_active_predictors = []
        self.best_predictor = Predictor()
        self.default_predictor = Predictor()
        self.default_predictor.condition = [-1]*J
        self.default_predictor.action = self.update_default_predictor()
        self.demand = 0.

    def update_default_predictor(self):
        a = np.sum([predictor.action[0]/predictor.actual_variance for predictor in self.predictors])
        b = np.sum([predictor.action[1]/predictor.actual_variance for predictor in self.predictors])
        variances = np.sum([1./predictor.actual_variance for predictor in self.predictors])
        return [a / variances, b / variances]

    def tournament(self, pool):
        challenger_1 = pool[random.randint(0, len(pool))]
        challenger_2 = pool[random.randint(0, len(pool))]
        if challenger_2.fitness > challenger_1.fitness:
            return challenger_2
        else:
            return challenger_1

        


########################################################################
######## MARKET ROUTINES ###############################################
########################################################################

def MA(n):
    if len(price) >= n:
        return np.mean(price[-n:])
    else:
        return np.mean(price)

def generate_market_state():
    # 1-6 indicators
    indicators_1 = [.25, .5, .75, .875, 1., 1.125]
    market_state[:6] = [1 if (price[-1]*(r/dividend[-1]) > indicator) else 0 for indicator in indicators_1]    
    # 7-10 indicators
    indicators_2 = [MA(5), MA(10), MA(100), MA(500)]
    market_state[6:10] = [1 if (price[-1] > indicator) else 0 for indicator in indicators_2]
    # 11 indicator fixed at 1
    market_state[10] = 1
    # 12 indicator fixed at 0 (from market_state initialization)

def store_data (t):
    if t == 0:
        for i,this_time_series in enumerate(time_series):
            f = open(time_series_labels[i] + '.csv', 'w')
            f.write(time_series_labels[i] + '\n')
            f.close()

    for i,this_time_series in enumerate(time_series):
        f = open(time_series_labels[i] + '.csv', 'a')
        f.write('%f\n' % this_time_series[-1])
        f.close()

################# INITIAL 500-STEP EVOLUTION ##################################
Investors = [] # investors' agentset

for i in range(N):
    investor = Investor()
    Investors.append(investor)


store_data(0)

for t in range(1,500):
    print t
    dividend.append(10. + 0.95*(dividend[-1] - 10.) + random.normal(0., np.sqrt(0.0743)))
    price.append(100.)
    volume.append(0.)
    hree_price.append(dividend[-1]/r)
    store_data(t)
    t += 1

################# ACTUAL TRADING  #############################################

for t in range(500,nrun):
    print t
    # dividend is posted
    dividend.append(10. + 0.95*(dividend[-1] - 10.) + random.normal(0., np.sqrt(0.0743)))
    # hree price is computed
    f = 0.95/(1 + r - 0.95)
    g = (1 + f)*(((1 - 0.95)*10. - gamma*0.0743)/r)
    hree_price.append(f*dividend[-1] + g)

    # investors activate predictors and pick most accurate
    generate_market_state()
    for investor in Investors:
        investor.active_predictors = [predictor for predictor in investor.predictors if predictor.compare_market_state]
        investor.active_predictors.append(investor.default_predictor)
        for predictor in investor.active_predictors:
            predictor.last_active = t
        np.random.shuffle(investor.active_predictors)
        investor.best_predictor = investor.active_predictors[np.argmin([predictor.variance for predictor in investor.active_predictors])]

    # price generation
    sum_1 = np.sum([investor.best_predictor.action[0]/(gamma*investor.best_predictor.variance) for investor in Investors])
    sum_2 = np.sum([investor.best_predictor.action[1]/(gamma*investor.best_predictor.variance) for investor in Investors])
    sum_3 = np.sum([1./(gamma*investor.best_predictor.variance) for investor in Investors])
    new_price = (S - dividend[-1]*sum_1 - sum_2)/(sum_1 - (1. + r)*sum_3)
    price.append(new_price)

    # transactions take place and volume is recorded
    day_volume = 0.
    for investor in Investors:
        demand = (investor.best_predictor.action[0]*(price[-1] + dividend[-1]) + investor.best_predictor.action[1] - price[-1]*(1. + r))/(gamma*investor.best_predictor.variance)
        day_volume += np.abs(demand - investor.demand)
        investor.demand = demand
    volume.append(0.5*day_volume)

    # predictors' updating
    for investor in Investors:
        for predictor in investor.last_active_predictors:
            squared_error = (price[-1] + dividend[-1] - predictor.action[0]*(price[-2] + dividend[-2]) - predictor.action[1])**2
            if squared_error > 500.:
                squared_error = 500.
            predictor.actual_variance = (1. - theta)*predictor.actual_variance + theta*squared_error
            predictor.fitness = - predictor.actual_variance - c*len([bit for bit in predictor.condition if bit != -1])
        investor.last_active_predictors = investor.active_predictors
        investor.active_predictors = []
        investor.update_default_predictor()

    # genetic algorithm
    for investor in Investors:    
        if random.random() < p_u:
            for predictor in investor.predictors:
                predictor.variance = predictor.actual_variance
                predictor.fitness = - predictor.variance - 0.005*len([bit for bit in predictor.condition if bit != -1])
            # elimination of 20% worst predictors
            investor.predictors = sorted(investor.predictors, key=lambda predictor: predictor.fitness)
            i = 0
            while (i+1 <= M/5.):
                del investor.predictors[0]
                i += 1
            # creation of new predictors
            elite_predictors = investor.predictors
            for j in range(i):
                new_predictor = Predictor()
                if (random.random() < p):
                    new_predictor.recombine(investor.tournament(elite_predictors), investor.tournament(elite_predictors), elite_predictors)
                else:
                    new_predictor.mutate(investor.tournament(elite_predictors))
                investor.predictors.append(predictor)
            if len(investor.predictors) != M:
                print 'Error: genetic evolution is not conservative. Number of predictors is now %d' % len(self.predictors)

    # generalization
    for investor in Investors:
        for predictor in investor.predictors:
            if t - predictor.last_active > 4000:
                predictor.mutate_for_inactivity()
                predictor.fitness = np.median([predictor.fitness for predictor in investor.predictors])

  
    store_data(t)
    
#plt.plot(price[500:])
#plt.plot(hree_price[500:], 'r--')
#plt.plot(volume[500:], 'b')
#plt.plot(np.array(hree_price[500:]) - np.array(price[500:]), 'r')
#plt.plot(action1, 'y--')
#plt.plot(action2, 'y')
#plt.plot(var, 'g')
#plt.show()


    

        
        
